from django.http import QueryDict
from django.test import RequestFactory

from debug_toolbar.panels.request import RequestPanel

from ..base import BaseTestCase

rf = RequestFactory()


class RequestPanelTestCase(BaseTestCase):
    panel_id = RequestPanel.panel_id

    def test_non_ascii_session(self):
        self.request.session = {"où": "où"}
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        self.assertIn("où", self.panel.content)

    def test_object_with_non_ascii_repr_in_request_params(self):
        request = rf.get("/non_ascii_request/")
        response = self.panel.process_request(request)
        self.panel.generate_stats(request, response)
        self.assertIn("nôt åscíì", self.panel.content)

    def test_insert_content(self):
        """
        Test that the panel only inserts content after generate_stats and
        not the process_request.
        """
        request = rf.get("/non_ascii_request/")
        response = self.panel.process_request(request)
        # ensure the panel does not have content yet.
        self.assertNotIn("nôt åscíì", self.panel.content)
        self.panel.generate_stats(request, response)
        # ensure the panel renders correctly.
        content = self.panel.content
        self.assertIn("nôt åscíì", content)
        self.assertValidHTML(content)

    def test_query_dict_for_request_in_method_get(self):
        """
        Test verifies the correctness of the statistics generation method
        in the case when the GET request is class QueryDict
        """
        self.request.GET = QueryDict("foo=bar")
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # ensure the panel GET request data is processed correctly.
        content = self.panel.content
        self.assertIn("foo", content)
        self.assertIn("bar", content)

    def test_dict_for_request_in_method_get(self):
        """
        Test verifies the correctness of the statistics generation method
        in the case when the GET request is class dict
        """
        self.request.GET = {"foo": "bar"}
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # ensure the panel GET request data is processed correctly.
        content = self.panel.content
        self.assertIn("foo", content)
        self.assertIn("bar", content)

    def test_query_dict_for_request_in_method_post(self):
        """
        Test verifies the correctness of the statistics generation method
        in the case when the POST request is class QueryDict
        """
        self.request.POST = QueryDict("foo=bar")
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # ensure the panel POST request data is processed correctly.
        content = self.panel.content
        self.assertIn("foo", content)
        self.assertIn("bar", content)

    def test_dict_for_request_in_method_post(self):
        """
        Test verifies the correctness of the statistics generation method
        in the case when the POST request is class dict
        """
        self.request.POST = {"foo": "bar"}
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # ensure the panel POST request data is processed correctly.
        content = self.panel.content
        self.assertIn("foo", content)
        self.assertIn("bar", content)

    def test_list_for_request_in_method_post(self):
        """
        Verify that the toolbar doesn't crash if request.POST contains unexpected data.

        See https://github.com/django-commons/django-debug-toolbar/issues/1621
        """
        self.request.POST = [{"a": 1}, {"b": 2}]
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        # ensure the panel POST request data is processed correctly.
        content = self.panel.content
        self.assertIn("[{&#x27;a&#x27;: 1}, {&#x27;b&#x27;: 2}]", content)

    def test_namespaced_url(self):
        request = rf.get("/admin/login/")
        response = self.panel.process_request(request)
        self.panel.generate_stats(request, response)
        panel_stats = self.panel.get_stats()
        self.assertEqual(panel_stats["view_urlname"], "admin:login")

    def test_session_list_sorted_or_not(self):
        """
        Verify the session is sorted when all keys are strings.

        See  https://github.com/django-commons/django-debug-toolbar/issues/1668
        """
        self.request.session = {
            1: "value",
            "data": ["foo", "bar", 1],
            (2, 3): "tuple_key",
        }
        data = {
            "list": [(1, "value"), ("data", ["foo", "bar", 1]), ((2, 3), "tuple_key")]
        }
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        panel_stats = self.panel.get_stats()
        self.assertEqual(panel_stats["session"], data)

        self.request.session = {
            "b": "b-value",
            "a": "a-value",
        }
        data = {"list": [("a", "a-value"), ("b", "b-value")]}
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)
        panel_stats = self.panel.get_stats()
        self.assertEqual(panel_stats["session"], data)

    def test_sensitive_post_data_sanitized(self):
        """Test that sensitive POST data is redacted."""
        self.request.POST = {"username": "testuser", "password": "secret123"}
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # Check that password is redacted in panel content
        content = self.panel.content
        self.assertIn("username", content)
        self.assertIn("testuser", content)
        self.assertIn("password", content)
        self.assertNotIn("secret123", content)
        self.assertIn("********************", content)

    def test_sensitive_get_data_sanitized(self):
        """Test that sensitive GET data is redacted."""
        self.request.GET = {"api_key": "abc123", "q": "search term"}
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # Check that api_key is redacted in panel content
        content = self.panel.content
        self.assertIn("api_key", content)
        self.assertNotIn("abc123", content)
        self.assertIn("********************", content)
        self.assertIn("q", content)
        self.assertIn("search term", content)

    def test_sensitive_cookie_data_sanitized(self):
        """Test that sensitive cookie data is redacted."""
        self.request.COOKIES = {"session_id": "abc123", "auth_token": "xyz789"}
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # Check that auth_token is redacted in panel content
        content = self.panel.content
        self.assertIn("session_id", content)
        self.assertIn("abc123", content)
        self.assertIn("auth_token", content)
        self.assertNotIn("xyz789", content)
        self.assertIn("********************", content)

    def test_sensitive_session_data_sanitized(self):
        """Test that sensitive session data is redacted."""
        self.request.session = {"user_id": 123, "auth_token": "xyz789"}
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # Check that auth_token is redacted in panel content
        content = self.panel.content
        self.assertIn("user_id", content)
        self.assertIn("123", content)
        self.assertIn("auth_token", content)
        self.assertNotIn("xyz789", content)
        self.assertIn("********************", content)

    def test_querydict_sanitized(self):
        """Test that sensitive data in QueryDict objects is properly redacted."""
        query_dict = QueryDict("username=testuser&password=secret123&token=abc456")
        self.request.GET = query_dict
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        # Check that sensitive data is redacted in panel content
        content = self.panel.content
        self.assertIn("username", content)
        self.assertIn("testuser", content)
        self.assertIn("password", content)
        self.assertNotIn("secret123", content)
        self.assertIn("token", content)
        self.assertNotIn("abc456", content)
        self.assertIn("********************", content)



import contextlib
import json

from django.http import HttpResponse, QueryDict
from django.http.request import HttpRequest, RawPostDataException
from django.template.loader import render_to_string
from django.templatetags.static import static
from django.urls import URLPattern, path
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from debug_toolbar.panels import Panel
from debug_toolbar.panels.history import views
from debug_toolbar.panels.history.forms import HistoryStoreForm


class HistoryPanel(Panel):
    """A panel to display History"""

    is_async = True
    title = _("History")
    nav_title = _("History")
    nav_icon = "debug_toolbar/img/history.svg"
    template = "debug_toolbar/panels/history.html"

    def get_headers(self, request: HttpRequest) -> dict:
        headers: dict = super().get_headers(request)
        observe_request = self.toolbar.get_observe_request()
        request_id = self.toolbar.request_id
        if request_id and observe_request(request):
            headers["djdt-request-id"] = request_id
        return headers

    @property
    def enabled(self) -> bool:
        # Do not show the history panel if the panels are rendered on request
        # rather than loaded via ajax.
        return super().enabled and not self.toolbar.should_render_panels()

    @property
    def is_historical(self) -> bool:
        """The HistoryPanel should not be included in the historical panels."""
        return False

    @classmethod
    def get_urls(cls) -> list[URLPattern]:
        return [
            path("history_sidebar/", views.history_sidebar, name="history_sidebar"),
            path("history_refresh/", views.history_refresh, name="history_refresh"),
        ]

    @property
    def nav_subtitle(self) -> str:
        return self.get_stats().get("request_url", "")

    @property
    def title_bar(self) -> str:
        """Render the refresh control shown in the panel's title bar."""
        return render_to_string(
            "debug_toolbar/panels/history_title_bar.html",
            {
                "refresh_form": HistoryStoreForm(
                    initial={
                        "request_id": self.toolbar.request_id,
                        "exclude_history": True,
                    }
                ),
            },
        )

    def generate_stats(self, request: HttpRequest, response: HttpResponse) -> None:
        data: QueryDict | None = None
        try:
            if request.method == "GET":
                data = request.GET.copy()
            else:
                data = request.POST.copy()
            # GraphQL tends to not be populated in POST. If the request seems
            # empty, check if it's a JSON request.
            if (
                not data
                and request.body
                and request.headers.get("content-type") == "application/json"
            ):
                with contextlib.suppress(ValueError):
                    data = json.loads(request.body)

        except RawPostDataException:
            # It is not guaranteed that we may read the request data (again).
            pass

        self.record_stats(
            {
                "request_url": request.get_full_path(),
                "request_method": request.method,
                "status_code": response.status_code,
                "data": data,
                "time": timezone.now(),
            }
        )

    @property
    def content(self) -> str:
        """Content of the panel when it's displayed in full screen.

        Fetch every store for the toolbar and include it in the template.
        """
        toolbar_history: dict[str, dict] = {}
        for request_id in reversed(self.toolbar.store.request_ids()):
            toolbar_history[request_id] = {
                "history_stats": self.toolbar.store.panel(
                    request_id, HistoryPanel.panel_id
                ),
                "form": HistoryStoreForm(
                    initial={"request_id": request_id, "exclude_history": True}
                ),
            }

        return render_to_string(
            self.template,
            {
                "current_request_id": self.toolbar.request_id,
                "toolbar_history": toolbar_history,
            },
        )

    @property
    def scripts(self) -> list[str]:
        scripts: list[str] = super().scripts
        scripts.append(static("debug_toolbar/js/history.js"))
        return scripts



import inspect
import mimetypes

from django.apps import AppConfig
from django.conf import settings
from django.core.checks import Error, Warning, register
from django.middleware.gzip import GZipMiddleware
from django.urls import NoReverseMatch, reverse
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from debug_toolbar import APP_NAME, settings as dt_settings
from debug_toolbar.settings import CONFIG_DEFAULTS


class DebugToolbarConfig(AppConfig):
    name = "debug_toolbar"
    verbose_name = _("Debug Toolbar")

    def ready(self):
        from debug_toolbar.toolbar import DebugToolbar

        # Import the panels when the app is ready and call their ready() methods.  This
        # allows panels like CachePanel to enable their instrumentation immediately.
        for cls in DebugToolbar.get_panel_classes():
            cls.ready()
        _manage_migrations_visibility(self.name)

    def import_models(self):
        """
        Override import models to avoid allowing users to dynamically fetch a model
        that doesn't have a table behind it unless they are using the DatabaseStore.
        This also prevents the command `migrate --run-syncdb` from creating tables
        for the models.
        """
        dt_config = dt_settings.get_config()
        if dt_config["TOOLBAR_STORE_CLASS"] == "debug_toolbar.store.DatabaseStore":
            return super().import_models()
        # Not using the database store, don't import the models
        self.models = {}


def _manage_migrations_visibility(app_name):
    """
    Adjust the toolbar's migration visibility by manipulating the
    project's settings.

    This is a hack since it's manipulating settings.
    """
    if (
        dt_settings.get_config()["TOOLBAR_STORE_CLASS"]
        != "debug_toolbar.store.DatabaseStore"
    ):
        # This effectively hides the migrations by telling Django they don't exist.
        settings.MIGRATION_MODULES.setdefault(app_name, None)


def check_template_config(config):
    """
    Checks if a template configuration is valid.

    The toolbar requires either the loaders to be unspecified or
    ``django.template.loaders.app_directories.Loader`` to be
    included in the loaders.
    If custom loaders are specified, then APP_DIRS must be True.
    """

    def flat_loaders(loaders):
        """
        Recursively flatten the settings list of template loaders.

        Check for (loader, [child_loaders]) tuples.
        Django's default cached loader uses this pattern.
        """
        for loader in loaders:
            if isinstance(loader, tuple):
                yield loader[0]
                yield from flat_loaders(loader[1])
            else:
                yield loader

    app_dirs = config.get("APP_DIRS", False)
    loaders = config.get("OPTIONS", {}).get("loaders", None)
    if loaders:
        loaders = list(flat_loaders(loaders))

    # By default the app loader is included.
    has_app_loaders = (
        loaders is None or "django.template.loaders.app_directories.Loader" in loaders
    )
    return has_app_loaders or app_dirs


@register
def check_middleware(app_configs, **kwargs):
    from debug_toolbar.middleware import DebugToolbarMiddleware

    errors = []
    gzip_index = None
    debug_toolbar_indexes = []

    if all(not check_template_config(config) for config in settings.TEMPLATES):
        errors.append(
            Warning(
                "At least one DjangoTemplates TEMPLATES configuration needs "
                "to use django.template.loaders.app_directories.Loader or "
                "have APP_DIRS set to True.",
                hint=(
                    "Include django.template.loaders.app_directories.Loader "
                    'in ["OPTIONS"]["loaders"]. Alternatively use '
                    "APP_DIRS=True for at least one "
                    "django.template.backends.django.DjangoTemplates "
                    "backend configuration."
                ),
                id="debug_toolbar.W006",
            )
        )

    # If old style MIDDLEWARE_CLASSES is being used, report an error.
    if settings.is_overridden("MIDDLEWARE_CLASSES"):
        errors.append(
            Warning(
                "debug_toolbar is incompatible with MIDDLEWARE_CLASSES setting.",
                hint="Use MIDDLEWARE instead of MIDDLEWARE_CLASSES",
                id="debug_toolbar.W004",
            )
        )
        return errors

    # Determine the indexes which gzip and/or the toolbar are installed at
    for i, middleware in enumerate(settings.MIDDLEWARE):
        if is_middleware_class(GZipMiddleware, middleware):
            gzip_index = i
        elif is_middleware_class(DebugToolbarMiddleware, middleware):
            debug_toolbar_indexes.append(i)

    if not debug_toolbar_indexes:
        # If the toolbar does not appear, report an error.
        errors.append(
            Warning(
                "debug_toolbar.middleware.DebugToolbarMiddleware is missing "
                "from MIDDLEWARE.",
                hint="Add debug_toolbar.middleware.DebugToolbarMiddleware to "
                "MIDDLEWARE.",
                id="debug_toolbar.W001",
            )
        )
    elif len(debug_toolbar_indexes) != 1:
        # If the toolbar appears multiple times, report an error.
        errors.append(
            Warning(
                "debug_toolbar.middleware.DebugToolbarMiddleware occurs "
                "multiple times in MIDDLEWARE.",
                hint="Load debug_toolbar.middleware.DebugToolbarMiddleware only "
                "once in MIDDLEWARE.",
                id="debug_toolbar.W002",
            )
        )
    elif gzip_index is not None and debug_toolbar_indexes[0] < gzip_index:
        # If the toolbar appears before the gzip index, report an error.
        errors.append(
            Warning(
                "debug_toolbar.middleware.DebugToolbarMiddleware occurs before "
                "django.middleware.gzip.GZipMiddleware in MIDDLEWARE.",
                hint="Move debug_toolbar.middleware.DebugToolbarMiddleware to "
                "after django.middleware.gzip.GZipMiddleware in MIDDLEWARE.",
                id="debug_toolbar.W003",
            )
        )
    return errors


@register
def check_panel_configs(app_configs, **kwargs):
    """Allow each panel to check the toolbar's integration for its own purposes."""
    from debug_toolbar.toolbar import DebugToolbar

    errors = []
    for panel_class in DebugToolbar.get_panel_classes():
        errors.extend(panel_class.run_checks())
    return errors


def is_middleware_class(middleware_class, middleware_path):
    try:
        middleware_cls = import_string(middleware_path)
    except ImportError:
        return
    return inspect.isclass(middleware_cls) and issubclass(
        middleware_cls, middleware_class
    )


@register
def check_panels(app_configs, **kwargs):
    errors = []
    panels = dt_settings.get_panels()
    if not panels:
        errors.append(
            Warning(
                "Setting DEBUG_TOOLBAR_PANELS is empty.",
                hint="Set DEBUG_TOOLBAR_PANELS to a non-empty list in your "
                "settings.py.",
                id="debug_toolbar.W005",
            )
        )
    return errors


@register
def js_mimetype_check(app_configs, **kwargs):
    """
    Check that JavaScript files are resolving to the correct content type.
    """
    # Ideally application/javascript is returned, but text/javascript is
    # acceptable.
    javascript_types = {"application/javascript", "text/javascript"}
    check_failed = not set(mimetypes.guess_type("toolbar.js")).intersection(
        javascript_types
    )
    if check_failed:
        return [
            Warning(
                "JavaScript files are resolving to the wrong content type.",
                hint="The Django Debug Toolbar may not load properly while mimetypes are misconfigured. "
                "See the Django documentation for an explanation of why this occurs.\n"
                "https://docs.djangoproject.com/en/stable/ref/contrib/staticfiles/#static-file-development-view\n"
                "\n"
                "This typically occurs on Windows machines. The suggested solution is to modify "
                "HKEY_CLASSES_ROOT in the registry to specify the content type for JavaScript "
                "files.\n"
                "\n"
                "[HKEY_CLASSES_ROOT\\.js]\n"
                '"Content Type"="application/javascript"',
                id="debug_toolbar.W007",
            )
        ]
    return []


@register
def debug_toolbar_installed_when_running_tests_check(app_configs, **kwargs):
    """
    Check that the toolbar is not being used when tests are running
    """
    # Check if show toolbar callback has changed
    show_toolbar_changed = (
        dt_settings.get_config()["SHOW_TOOLBAR_CALLBACK"]
        != CONFIG_DEFAULTS["SHOW_TOOLBAR_CALLBACK"]
    )

    if not hasattr(settings, "ROOT_URLCONF"):
        # reverse will raise AttributeError if ROOT_URLCONF is not defined
        toolbar_urls_installed = False
    else:
        try:
            # Check if the toolbar's urls are installed
            reverse(f"{APP_NAME}:render_panel")
            toolbar_urls_installed = True
        except NoReverseMatch:
            toolbar_urls_installed = False

    # If the user is using the default SHOW_TOOLBAR_CALLBACK,
    # then the middleware will respect the change to settings.DEBUG.
    # However, if the user has changed the callback to:
    # DEBUG_TOOLBAR_CONFIG = {"SHOW_TOOLBAR_CALLBACK": lambda request: DEBUG}
    # where DEBUG is not settings.DEBUG, then it won't pick up that Django'
    # test runner has changed the value for settings.DEBUG, and the middleware
    # will inject the toolbar, while the URLs aren't configured leading to a
    # NoReverseMatch error.
    likely_error_setup = show_toolbar_changed and not toolbar_urls_installed

    if (
        not settings.DEBUG
        and dt_settings.get_config()["IS_RUNNING_TESTS"]
        and likely_error_setup
    ):
        return [
            Error(
                "The Django Debug Toolbar can't be used with tests",
                hint="Django changes the DEBUG setting to False when running "
                "tests. By default the Django Debug Toolbar is installed because "
                "DEBUG is set to True. For most cases, you need to avoid installing "
                "the toolbar when running tests. If you feel this check is in error, "
                "you can set `DEBUG_TOOLBAR_CONFIG['IS_RUNNING_TESTS'] = False` to "
                "bypass this check.",
                id="debug_toolbar.E001",
            )
        ]
    else:
        return []


@register
def check_settings(app_configs, **kwargs):
    errors = []
    USER_CONFIG = getattr(settings, "DEBUG_TOOLBAR_CONFIG", {})
    if "OBSERVE_REQUEST_CALLBACK" in USER_CONFIG:
        errors.append(
            Warning(
                "The deprecated OBSERVE_REQUEST_CALLBACK setting is present in DEBUG_TOOLBAR_CONFIG.",
                hint="Use the UPDATE_ON_FETCH and/or SHOW_TOOLBAR_CALLBACK settings instead.",
                id="debug_toolbar.W008",
            )
        )
    return errors



from django.contrib import admin
from django.contrib.auth.views import LoginView
from django.urls import include, path, re_path

from . import views
from .models import NonAsciiRepr

urlpatterns = [
    re_path(
        r"^resolving1/(.+)/(.+)/$", views.resolving_view, name="positional-resolving"
    ),
    path("resolving2/<path:arg1>/<path:arg2>/", views.resolving_view),
    re_path(r"^resolving3/(.+)/$", views.resolving_view, {"arg2": "default"}),
    re_path(r"^regular/(?P<title>.*)/$", views.regular_view),
    re_path(r"^template_response/(?P<title>.*)/$", views.template_response_view),
    re_path(r"^regular_jinja/(?P<title>.*)/$", views.regular_jinjia_view),
    path("non_ascii_request/", views.regular_view, {"title": NonAsciiRepr()}),
    path("new_user/", views.new_user),
    path("execute_sql/", views.execute_sql),
    path("execute_json_sql/", views.execute_json_sql),
    path("execute_union_sql/", views.execute_union_sql),
    path("execute_binary_sql/", views.execute_binary_sql),
    path("async_execute_sql/", views.async_execute_sql),
    path("async_execute_json_sql/", views.async_execute_json_sql),
    path("async_execute_union_sql/", views.async_execute_union_sql),
    path("async_execute_sql_concurrently/", views.async_execute_sql_concurrently),
    path("cached_view/", views.cached_view),
    path("cached_low_level_view/", views.cached_low_level_view),
    path("cache_with_non_json_key/", views.cache_with_non_json_key_view),
    path("json_view/", views.json_view),
    path("redirect/", views.redirect_view),
    path("ajax/", views.ajax_view),
    path("login_without_redirect/", LoginView.as_view(redirect_field_name=None)),
    path("csp_view/", views.csp_view),
    path("admin/", admin.site.urls),
    path("__debug__/", include("debug_toolbar.urls")),
]



from django import VERSION
from django.utils.translation import gettext_lazy as _, ngettext

from debug_toolbar.panels import Panel

if VERSION >= (6, 0):
    from django.tasks.signals import task_enqueued
else:
    task_enqueued = None


class TasksPanel(Panel):
    """
    Panel that displays the tasks queued during the request.

    This relies on Django's built-in tasks framework (``django.tasks``),
    which was added in Django 6.0. On older versions of Django, the panel
    explains that upgrading Django is required in order to see task
    information.
    """

    template = "debug_toolbar/panels/tasks.html"

    is_async = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tasks: list = []  # populated with TaskResult instances

    nav_icon = "debug_toolbar/img/tasks.svg"

    nav_title = _("Tasks")

    @property
    def nav_subtitle(self):
        if VERSION < (6, 0):
            return _("Requires Django 6.0+")
        tasks = self.get_stats()["tasks"]
        return ngettext("%(count)d task", "%(count)d tasks", len(tasks)) % {
            "count": len(tasks)
        }

    title = _("Tasks")

    def _record_task(self, sender, task_result, **kwargs):
        self.tasks.append(task_result)

    def enable_instrumentation(self):
        if task_enqueued is not None:
            task_enqueued.connect(self._record_task)

    def disable_instrumentation(self):
        if task_enqueued is not None:
            task_enqueued.disconnect(self._record_task)

    def generate_stats(self, request, response):
        # TaskResult and Task aren't JSON-serializable which the toolbar's
        # storage mechanism requires
        tasks = [
            {
                "task": {
                    "module_path": task_result.task.module_path,
                    "queue_name": task_result.task.queue_name,
                    "priority": task_result.task.priority,
                    "run_after": task_result.task.run_after,
                },
                "backend": task_result.backend,
                "status": task_result.status,
                "args": task_result.args,
                "kwargs": task_result.kwargs,
            }
            for task_result in self.tasks
        ]
        self.record_stats(
            {
                "tasks_available": VERSION >= (6, 0),
                "tasks": tasks,
            }
        )



#!/usr/bin/env python3

import sys

sys.stderr.write(
    """\
===============================
Unsupported installation method
===============================
This project no longer supports installation with `python setup.py install`.
Please use `python -m pip install .` instead.
"""
)
sys.exit(1)



import asyncio

from asgiref.sync import sync_to_async
from django.contrib.auth.models import User
from django.core.cache import cache
from django.http import HttpResponseRedirect, JsonResponse
from django.shortcuts import render
from django.template.response import TemplateResponse
from django.views.decorators.cache import cache_page

from debug_toolbar.utils import get_csp_nonce
from tests.models import Binary, PostgresJSON


def execute_sql(request):
    list(User.objects.all())
    return render(request, "base.html")


def execute_binary_sql(request):
    list(Binary.objects.filter(field=b"\x01\x02\x03"))
    return render(request, "base.html")


def execute_json_sql(request):
    list(PostgresJSON.objects.filter(field__contains={"foo": "bar"}))
    return render(request, "base.html")


async def async_execute_json_sql(request):
    list_store = []
    # make async query with filter, which is compatible with async for.
    async for obj in PostgresJSON.objects.filter(field__contains={"foo": "bar"}):
        list_store.append(obj)
    return render(request, "base.html")


def execute_union_sql(request):
    list(User.objects.all().union(User.objects.all(), all=True))
    return render(request, "base.html")


async def async_execute_union_sql(request):
    list_store = []
    # make async query with filter, which is compatible with async for.
    users = User.objects.all().union(User.objects.all(), all=True)
    async for user in users:
        list_store.append(user)
    return render(request, "base.html")


async def async_execute_sql(request):
    """
    Some query API can be executed asynchronously but some requires
    async version of itself.

    https://docs.djangoproject.com/en/5.1/topics/db/queries/#asynchronous-queries
    """
    list_store = []

    # make async query with filter, which is compatible with async for.
    async for user in User.objects.filter(username="test"):
        list_store.append(user)

    # make async query with afirst
    async_fetched_user = await User.objects.filter(username="test").afirst()
    list_store.append(async_fetched_user)
    return render(request, "base.html")


async def async_execute_sql_concurrently(request):
    await asyncio.gather(sync_to_async(list)(User.objects.all()), User.objects.acount())
    return render(request, "base.html")


def regular_view(request, title):
    return render(request, "basic.html", {"title": title})


def csp_view(request):
    """Use request.csp_nonce to inject it into the headers"""
    nonce = get_csp_nonce(request)
    return render(request, "basic.html", {"title": f"CSP {nonce}"})


def template_response_view(request, title):
    return TemplateResponse(request, "basic.html", {"title": title})


def new_user(request, username="joe"):
    User.objects.create_user(username=username)
    return render(request, "basic.html", {"title": "new user"})


def resolving_view(request, arg1, arg2):
    # see test_url_resolving in tests.py
    return render(request, "base.html")


@cache_page(60)
def cached_view(request):
    return render(request, "base.html")


def cached_low_level_view(request):
    key = "spam"
    value = cache.get(key)
    if not value:
        value = "eggs"
        cache.set(key, value, 60)
    return render(request, "base.html")


def cache_with_non_json_key_view(request):
    cache.set_many({str: "this-is-a-string", "foo": "bar"})
    return render(request, "base.html")


def json_view(request):
    return JsonResponse({"foo": "bar"})


def regular_jinjia_view(request, title):
    return render(request, "basic.jinja", {"title": title}, using="jinja2")


def listcomp_view(request):
    lst = [i for i in range(50000) if i % 2 == 0]
    return render(request, "basic.html", {"title": "List comprehension", "lst": lst})


def redirect_view(request):
    return HttpResponseRedirect("/regular/redirect/")


def ajax_view(request):
    return render(request, "ajax/ajax.html")



from django.test import RequestFactory, override_settings

from debug_toolbar.panels.settings import SettingsPanel

from ..base import BaseTestCase, IntegrationTestCase

rf = RequestFactory()


class SettingsPanelTestCase(BaseTestCase):
    panel_id = SettingsPanel.panel_id

    def test_panel_recording(self):
        self.request = rf.post("/", data={"foo": "bar"})
        response = self.panel.process_request(self.request)
        self.panel.generate_stats(self.request, response)

        settings = self.panel.get_stats()["settings"]
        self.assertEqual(settings["USE_THOUSAND_SEPARATOR"], "False")
        self.assertEqual(settings["ABSOLUTE_URL_OVERRIDES"], "{}")
        self.assertEqual(settings["EMAIL_HOST"], "'localhost'")
        self.assertEqual(settings["EMAIL_PORT"], "25")


@override_settings(DEBUG=True)
class SettingsIntegrationTestCase(IntegrationTestCase):
    def test_panel_title(self):
        response = self.client.get("/regular/basic/")
        # The settings module is None due to using Django's UserSettingsHolder
        # in tests.
        self.assertContains(
            response,
            '<a href="#" title="Settings from None" class="SettingsPanel"'
            ' aria-controls="SettingsPanel" aria-expanded="false">',
        )
        self.assertContains(
            response,
            '<svg class="djdt-nav-icon" aria-hidden="true" focusable="false"',
        )
        self.assertNotContains(response, 'src="/static/debug_toolbar/img/settings.svg"')
        self.assertContains(
            response,
            """
            <div id="SettingsPanel" class="djdt-panelContent djdt-hidden">
            <div class="djDebugPanelTitle">
            <h3>Settings from None</h3>
            <button type="button" class="djDebugClose">×</button>
            </div>
            <div class="djDebugPanelContent">
            <div class="djdt-loader"></div>
            <div class="djdt-scroll" tabindex="0"></div>
            </div>
            </div>
            """,
            html=True,
        )



from functools import cache, lru_cache
from html import escape
from itertools import cycle

import sqlparse
from django.dispatch import receiver
from django.test.signals import setting_changed
from sqlparse import tokens as T
from sqlparse.exceptions import SQLParseError

from debug_toolbar import settings as dt_settings


class ElideSelectListsFilter:
    """sqlparse filter to elide the select list from top-level SELECT ... FROM clauses,
    if present"""

    def process(self, stream):
        allow_elision = True
        for token_type, value in stream:
            yield token_type, value
            if token_type in T.Keyword:
                keyword = value.upper()
                if allow_elision and keyword == "SELECT":
                    yield from self.elide_until_from(stream)
                allow_elision = keyword in ["EXCEPT", "INTERSECT", "UNION"]

    @staticmethod
    def elide_until_from(stream):
        has_dot = False
        saved_tokens = []
        for token_type, value in stream:
            if token_type in T.Keyword and value.upper() == "FROM":
                # Do not elide a select lists that do not contain dots (used to separate
                # table names from column names) in order to preserve
                #    SELECT COUNT(*) AS `__count` FROM ...
                # and
                #    SELECT (1) AS `a` FROM ...
                # queries.
                if not has_dot:
                    yield from saved_tokens
                else:
                    # U+2022: Unicode character 'BULLET'
                    yield T.Other, " \u2022\u2022\u2022 "
                yield token_type, value
                break
            if not has_dot:
                if token_type in T.Punctuation and value == ".":
                    has_dot = True
                else:
                    saved_tokens.append((token_type, value))


class BoldKeywordFilter:
    """sqlparse filter to bold SQL keywords"""

    def process(self, stmt):
        idx = 0
        while idx < len(stmt.tokens):
            token = stmt[idx]
            if token.is_keyword:
                stmt.insert_before(idx, sqlparse.sql.Token(T.Other, "<strong>"))
                stmt.insert_after(
                    idx + 1,
                    sqlparse.sql.Token(T.Other, "</strong>"),
                    skip_ws=False,
                )
                idx += 2
            elif token.is_group:
                self.process(token)
            idx += 1


def escaped_value(token):
    # Don't escape T.Whitespace tokens because AlignedIndentFilter inserts its tokens as
    # T.Whitesapce, and in our case those tokens are actually HTML.
    if token.ttype in (T.Other, T.Whitespace):
        return token.value
    return escape(token.value, quote=False)


class EscapedStringSerializer:
    """sqlparse post-processor to convert a Statement into a string escaped for
    inclusion in HTML ."""

    @staticmethod
    def process(stmt):
        return "".join(escaped_value(token) for token in stmt.flatten())


def reformat_sql(sql, *, with_toggle=False):
    formatted = parse_sql(sql)
    if not with_toggle:
        return formatted
    simplified = parse_sql(sql, simplify=True)

    uncollapsed = f'<span class="djDebugUncollapsed">{simplified}</span>'
    collapsed = f'<span class="djDebugCollapsed djdt-hidden">{formatted}</span>'
    return collapsed + uncollapsed


@lru_cache(maxsize=128)
def parse_sql(sql, *, simplify=False):
    stack = get_filter_stack(simplify=simplify)
    try:
        return "".join(stack.run(sql))
    except SQLParseError:
        # The query either exceeds the number of tokens or depth of tokens.
        # Recreate the FilterStack and explicitly disable the grouping to avoid
        # those errors.
        stack = get_filter_stack(simplify=simplify)
        stack._grouping = False
        return "".join(stack.run(sql))


@cache
def get_filter_stack(*, simplify):
    stack = sqlparse.engine.FilterStack()
    if simplify:
        stack.preprocess.append(ElideSelectListsFilter())
    else:
        if dt_settings.get_config()["PRETTIFY_SQL"]:
            stack.enable_grouping()
        stack.stmtprocess.append(
            sqlparse.filters.AlignedIndentFilter(char="&nbsp;", n="<br/>")
        )
    stack.stmtprocess.append(BoldKeywordFilter())
    stack.postprocess.append(EscapedStringSerializer())  # Statement -> str
    return stack


@receiver(setting_changed)
def clear_caches(*, setting, **kwargs):
    if setting == "DEBUG_TOOLBAR_CONFIG":
        parse_sql.cache_clear()
        get_filter_stack.cache_clear()


def contrasting_color_generator():
    return cycle(
        [
            "#0C375A",
            "#21A0A0",
            "#FFC300",
            "#FF5733",
            "#C70039",
            "#900C3F",
            "#581845",
            "#F1C40F",
        ]
    )



from time import perf_counter

from django.template.loader import render_to_string
from django.templatetags.static import static
from django.utils.translation import gettext_lazy as _

from debug_toolbar.panels import Panel

try:
    import resource  # Not available on Win32 systems
except ImportError:
    resource = None


class TimerPanel(Panel):
    """
    Panel that displays the time a response took in milliseconds.
    """

    is_async = True

    def nav_subtitle(self):
        stats = self.get_stats()
        if stats.get("utime"):
            utime = stats.get("utime")
            stime = stats.get("stime")
            return _("CPU: %(cum)0.2fms (%(total)0.2fms)") % {
                "cum": (utime + stime),
                "total": stats["total_time"],
            }
        elif "total_time" in stats:
            return _("Total: %0.2fms") % stats["total_time"]
        else:
            return ""

    has_content = resource is not None

    title = _("Time")
    nav_icon = "debug_toolbar/img/timer.svg"

    template = "debug_toolbar/panels/timer.html"

    @property
    def content(self):
        stats = self.get_stats()
        rows = (
            (_("User CPU time"), _("%(utime)0.3f ms") % stats),
            (_("System CPU time"), _("%(stime)0.3f ms") % stats),
            (_("Total CPU time"), _("%(total)0.3f ms") % stats),
            (_("Elapsed time"), _("%(total_time)0.3f ms") % stats),
            (
                _("Including DebugToolbar time"),
                _("%(toolbar_time)0.3f ms") % stats,
            ),
            (
                _("Context switches"),
                _("%(vcsw)d voluntary, %(ivcsw)d involuntary") % stats,
            ),
        )
        return render_to_string(self.template, {"rows": rows})

    @property
    def scripts(self):
        scripts = super().scripts
        scripts.append(static("debug_toolbar/js/timer.js"))
        return scripts

    def process_request(self, request):
        self._start_time = perf_counter()
        if self.has_content:
            self._start_rusage = resource.getrusage(resource.RUSAGE_SELF)
        return super().process_request(request)

    def serialize_rusage(self, data):
        fields_to_serialize = [
            "ru_utime",
            "ru_stime",
            "ru_nvcsw",
            "ru_nivcsw",
            "ru_minflt",
            "ru_majflt",
        ]
        return {field: getattr(data, field) for field in fields_to_serialize}

    def generate_stats(self, request, response):
        stats = {}
        if hasattr(self, "_start_time"):
            stats["total_time"] = (perf_counter() - self._start_time) * 1000
        if self.has_content:
            self._end_rusage = resource.getrusage(resource.RUSAGE_SELF)
            start = self.serialize_rusage(self._start_rusage)
            end = self.serialize_rusage(self._end_rusage)
            stats.update(
                {
                    "utime": 1000 * self._elapsed_ru(start, end, "ru_utime"),
                    "stime": 1000 * self._elapsed_ru(start, end, "ru_stime"),
                    "vcsw": self._elapsed_ru(start, end, "ru_nvcsw"),
                    "ivcsw": self._elapsed_ru(start, end, "ru_nivcsw"),
                    "minflt": self._elapsed_ru(start, end, "ru_minflt"),
                    "majflt": self._elapsed_ru(start, end, "ru_majflt"),
                }
            )
            stats["total"] = stats["utime"] + stats["stime"]
            # these are documented as not meaningful under Linux.  If you're
            # running BSD feel free to enable them, and add any others that I
            # hadn't gotten to before I noticed that I was getting nothing but
            # zeroes and that the docs agreed. :-(
            #
            #        stats['blkin'] = self._elapsed_ru(start, end, 'ru_inblock')
            #        stats['blkout'] = self._elapsed_ru(start, end, 'ru_oublock')
            #        stats['swap'] = self._elapsed_ru(start, end, 'ru_nswap')
            #        stats['rss'] = self._end_rusage.ru_maxrss
            #        stats['srss'] = self._end_rusage.ru_ixrss
            #        stats['urss'] = self._end_rusage.ru_idrss
            #        stats['usrss'] = self._end_rusage.ru_isrss

        if hasattr(self, "_toolbar_start_time"):
            stats["toolbar_time"] = (perf_counter() - self._toolbar_start_time) * 1000

        self.record_stats(stats)

    def generate_server_timing(self, request, response):
        stats = self.get_stats()

        self.record_server_timing("utime", "User CPU time", stats.get("utime", 0))
        self.record_server_timing("stime", "System CPU time", stats.get("stime", 0))
        self.record_server_timing("total", "Total CPU time", stats.get("total", 0))
        self.record_server_timing(
            "total_time", "Elapsed time", stats.get("total_time", 0)
        )
        self.record_server_timing(
            "toolbar_time", "Including DebugToolbar time", stats.get("toolbar_time", 0)
        )

    @staticmethod
    def _elapsed_ru(start, end, name):
        return end.get(name) - start.get(name)

    def enable_instrumentation(self):
        self._toolbar_start_time = perf_counter()



from django.db import models
from django.utils.translation import gettext_lazy as _


class HistoryEntry(models.Model):
    request_id = models.UUIDField(primary_key=True)
    data = models.JSONField(default=dict)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        verbose_name = _("history entry")
        verbose_name_plural = _("history entries")
        ordering = ["-created_at"]

    def __str__(self):
        return str(self.request_id)



import base64
import contextlib
import contextvars
import datetime
from time import perf_counter

import django.test.testcases
from django.apps import apps
from django.core.exceptions import ImproperlyConfigured

from debug_toolbar import settings as dt_settings
from debug_toolbar.sanitize import force_str
from debug_toolbar.utils import get_stack_trace, get_template_info

Psycopg3Binary = None

try:
    import psycopg

    PostgresJson = psycopg.types.json.Jsonb
    STATUS_IN_TRANSACTION = psycopg.pq.TransactionStatus.INTRANS
    Psycopg3Binary = psycopg.Binary
except ImportError:
    try:
        from psycopg2._json import Json as PostgresJson
        from psycopg2.extensions import STATUS_IN_TRANSACTION
    except ImportError:
        PostgresJson = None
        STATUS_IN_TRANSACTION = None

try:
    from psycopg2.extensions import Binary as Psycopg2Binary
except ImportError:
    Psycopg2Binary = None

_PostGISAdapter = None
try:
    from django.contrib.gis.db.backends.postgis.adapter import (
        PostGISAdapter as _PostGISAdapter,
    )
except (ImportError, ImproperlyConfigured):
    pass

# Prevents SQL queries from being sent to the DB. It's used
# by the TemplatePanel to prevent the toolbar from issuing
# additional queries.
allow_sql = contextvars.ContextVar("debug-toolbar-allow-sql", default=True)


DDT_MODELS = {
    m._meta.db_table for m in apps.get_app_config("debug_toolbar").get_models()
}


class SQLQueryTriggered(Exception):
    """Thrown when template panel triggers a query"""


def wrap_cursor(connection):
    # When running a SimpleTestCase, Django monkey patches some DatabaseWrapper
    # methods, including .cursor() and .chunked_cursor(), to raise an exception
    # if the test code tries to access the database, and then undoes the monkey
    # patching when the test case is finished.  If we monkey patch those methods
    # also, Django's process of undoing those monkey patches will fail.  To
    # avoid this failure, and because database access is not allowed during a
    # SimpleTestCase anyway, skip applying our instrumentation monkey patches if
    # we detect that Django has already monkey patched DatabaseWrapper.cursor().
    if isinstance(connection.cursor, django.test.testcases._DatabaseFailure):
        return
    if not hasattr(connection, "_djdt_cursor"):
        connection._djdt_cursor = connection.cursor
        connection._djdt_chunked_cursor = connection.chunked_cursor
        connection._djdt_logger = None

        def cursor(*args, **kwargs):
            # Per the DB API cursor() does not accept any arguments. There's
            # some code in the wild which does not follow that convention,
            # so we pass on the arguments even though it's not clean.
            # See:
            # https://github.com/django-commons/django-debug-toolbar/pull/615
            # https://github.com/django-commons/django-debug-toolbar/pull/896
            logger = connection._djdt_logger
            cursor = connection._djdt_cursor(*args, **kwargs)
            if logger is None:
                return cursor
            mixin = NormalCursorMixin if allow_sql.get() else ExceptionCursorMixin
            return patch_cursor_wrapper_with_mixin(cursor.__class__, mixin)(
                cursor.cursor, connection, logger
            )

        def chunked_cursor(*args, **kwargs):
            # prevent double wrapping
            # solves https://github.com/django-commons/django-debug-toolbar/issues/1239
            logger = connection._djdt_logger
            cursor = connection._djdt_chunked_cursor(*args, **kwargs)
            if logger is not None and not isinstance(cursor, DjDTCursorWrapperMixin):
                mixin = NormalCursorMixin if allow_sql.get() else ExceptionCursorMixin
                return patch_cursor_wrapper_with_mixin(cursor.__class__, mixin)(
                    cursor.cursor, connection, logger
                )
            return cursor

        connection.cursor = cursor
        connection.chunked_cursor = chunked_cursor


def patch_cursor_wrapper_with_mixin(base_wrapper, mixin):
    class DjDTCursorWrapper(mixin, base_wrapper):
        pass

    return DjDTCursorWrapper


class DjDTCursorWrapperMixin:
    def __init__(self, cursor, db, logger):
        super().__init__(cursor, db)
        # logger must implement a ``record`` method
        self.logger = logger


class ExceptionCursorMixin(DjDTCursorWrapperMixin):
    """
    Wraps a cursor and raises an exception on any operation.
    Used in Templates panel.
    """

    def __getattr__(self, attr):
        raise SQLQueryTriggered()


class NormalCursorMixin(DjDTCursorWrapperMixin):
    """
    Wraps a cursor and logs queries.
    """

    def _decode(self, param):
        if PostgresJson and isinstance(param, PostgresJson):
            # psycopg3
            if hasattr(param, "obj"):
                return param.dumps(param.obj)
            # psycopg2
            if hasattr(param, "adapted"):
                return param.dumps(param.adapted)

        # If a sequence type, decode each element separately
        if isinstance(param, (tuple, list)):
            return [self._decode(element) for element in param]

        # If a dictionary type, decode each value separately
        if isinstance(param, dict):
            return {key: self._decode(value) for key, value in param.items()}

        # GeoDjango PostGIS geometry parameters: extract EWKB bytes and metadata
        # so the adapter can be reconstructed on the way back out for SELECT/EXPLAIN.
        if _PostGISAdapter is not None and isinstance(param, _PostGISAdapter):
            return {
                "__djdt_postgis__": base64.b64encode(param.ewkb).decode("ascii"),
                "is_geometry": param.is_geometry,
                "geography": param.geography,
            }

        # Binary data is handled by DebugToolbarJSONEncoder in store.py.
        # Django's BinaryField calls connection.Database.Binary() which wraps
        # bytes in a driver-specific adapter: psycopg2 uses .adapted, psycopg3
        # uses .obj. memoryview: psycopg2/sqlite3 binary column values.
        if isinstance(param, (bytes, bytearray, memoryview)):
            return bytes(param)
        if Psycopg2Binary is not None and isinstance(param, Psycopg2Binary):
            return bytes(param.adapted)
        if Psycopg3Binary is not None and isinstance(param, Psycopg3Binary):
            return bytes(param.obj)

        # make sure datetime, date and time are converted to string by force_str
        CONVERT_TYPES = (datetime.datetime, datetime.date, datetime.time)
        return force_str(param, strings_only=not isinstance(param, CONVERT_TYPES))

    def _last_executed_query(self, sql, params):
        """Get the last executed query from the connection."""
        # Django's psycopg3 backend creates a new cursor in its implementation of the
        # .last_executed_query() method.  To avoid wrapping that cursor, temporarily set
        # the DatabaseWrapper's ._djdt_logger attribute to None.  This will cause the
        # monkey-patched .cursor() and .chunked_cursor() methods to skip the wrapping
        # process during the .last_executed_query() call.
        self.db._djdt_logger = None
        try:
            return self.db.ops.last_executed_query(self.cursor, sql, params)
        finally:
            self.db._djdt_logger = self.logger

    def _record(self, method, sql, params):
        alias = self.db.alias
        vendor = self.db.vendor

        if vendor == "postgresql":
            # The underlying DB connection (as opposed to Django's wrapper)
            conn = self.db.connection
            initial_conn_status = conn.info.transaction_status

        start_time = perf_counter()
        try:
            return method(sql, params)
        finally:
            stop_time = perf_counter()
            duration = (stop_time - start_time) * 1000
            _params = None
            with contextlib.suppress(TypeError):
                # Decode params - binary data will be handled by DebugToolbarJSONEncoder
                # in store.py when the panel data is serialized
                _params = self._decode(params)
            template_info = get_template_info()

            # Sql might be an object (such as psycopg Composed).
            # For logging purposes, make sure it's str.
            if vendor == "postgresql" and not isinstance(sql, str):
                if isinstance(sql, bytes):
                    sql = sql.decode("utf-8")
                else:
                    sql = sql.as_string(conn)
            else:
                sql = str(sql)

            kwargs = {
                "vendor": vendor,
                "alias": alias,
                "sql": self._last_executed_query(sql, params),
                "duration": duration,
                "raw_sql": sql,
                "params": _params,
                "stacktrace": get_stack_trace(skip=2),
                "template_info": template_info,
            }

            if vendor == "postgresql":
                # If an erroneous query was ran on the connection, it might
                # be in a state where checking isolation_level raises an
                # exception.
                try:
                    iso_level = conn.isolation_level
                except conn.InternalError:
                    iso_level = "unknown"
                # PostgreSQL does not expose any sort of transaction ID, so it is
                # necessary to generate synthetic transaction IDs here.  If the
                # connection was not in a transaction when the query started, and was
                # after the query finished, a new transaction definitely started, so get
                # a new transaction ID from logger.new_transaction_id().  If the query
                # was in a transaction both before and after executing, make the
                # assumption that it is the same transaction and get the current
                # transaction ID from logger.current_transaction_id().  There is an edge
                # case where Django can start a transaction before the first query
                # executes, so in that case logger.current_transaction_id() will
                # generate a new transaction ID since one does not already exist.
                final_conn_status = conn.info.transaction_status
                if final_conn_status == STATUS_IN_TRANSACTION:
                    if initial_conn_status == STATUS_IN_TRANSACTION:
                        trans_id = self.logger.current_transaction_id(alias)
                    else:
                        trans_id = self.logger.new_transaction_id(alias)
                else:
                    trans_id = None

                kwargs.update(
                    {
                        "trans_id": trans_id,
                        "trans_status": conn.info.transaction_status,
                        "iso_level": iso_level,
                    }
                )

            # Skip tracking for toolbar models by default.
            # This can be overridden by setting SKIP_TOOLBAR_QUERIES = False
            if not dt_settings.get_config()["SKIP_TOOLBAR_QUERIES"] or not any(
                table in sql for table in DDT_MODELS
            ):
                # We keep `sql` to maintain backwards compatibility
                self.logger.record(**kwargs)

    def callproc(self, procname, params=None):
        return self._record(super().callproc, procname, params)

    def execute(self, sql, params=None):
        return self._record(super().execute, sql, params)

    def executemany(self, sql, param_list):
        return self._record(super().executemany, sql, param_list)



from html.parser import HTMLParser

from django.utils.translation import gettext_lazy as _

from debug_toolbar.panels import Panel
from debug_toolbar.utils import is_processable_html_response


class FormParser(HTMLParser):
    """
    HTML form parser, used to check for invalid configurations of forms that
    take file inputs.
    """

    def __init__(self):
        super().__init__()
        self.in_form = False
        self.current_form = {}
        self.forms = []
        self.form_ids = []
        self.referenced_file_inputs = []

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if tag == "form":
            self.in_form = True
            form_id = attrs.get("id")
            if form_id:
                self.form_ids.append(form_id)
            self.current_form = {
                "file_form": False,
                "form_attrs": attrs,
                "submit_element_attrs": [],
            }
        elif (
            self.in_form
            and tag == "input"
            and attrs.get("type") == "file"
            and (not attrs.get("form") or attrs.get("form") == "")
        ):
            self.current_form["file_form"] = True
        elif (
            self.in_form
            and (
                (tag == "input" and attrs.get("type") in {"submit", "image"})
                or tag == "button"
            )
            and (not attrs.get("form") or attrs.get("form") == "")
        ):
            self.current_form["submit_element_attrs"].append(attrs)
        elif tag == "input" and attrs.get("form"):
            self.referenced_file_inputs.append(attrs)

    def handle_endtag(self, tag):
        if tag == "form" and self.in_form:
            self.forms.append(self.current_form)
            self.in_form = False


class AlertsPanel(Panel):
    """
    A panel to alert users to issues.
    """

    messages = {
        "form_id_missing_enctype": _(
            'Form with id "{form_id}" contains file input, but does not have the attribute enctype="multipart/form-data".'
        ),
        "form_missing_enctype": _(
            'Form contains file input, but does not have the attribute enctype="multipart/form-data".'
        ),
        "input_refs_form_missing_enctype": _(
            'Input element references form with id "{form_id}", but the form does not have the attribute enctype="multipart/form-data".'
        ),
    }

    title = _("Alerts")
    nav_icon = "debug_toolbar/img/alerts.svg"

    is_async = True

    template = "debug_toolbar/panels/alerts.html"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alerts = []

    @property
    def nav_subtitle(self):
        if alerts := self.get_stats().get("alerts"):
            alert_text = "alert" if len(alerts) == 1 else "alerts"
            return f"{len(alerts)} {alert_text}"
        else:
            return ""

    def add_alert(self, alert):
        self.alerts.append(alert)

    def check_invalid_file_form_configuration(self, html_content):
        """
        Inspects HTML content for a form that includes a file input but does
        not have the encoding type set to multipart/form-data, and warns the
        user if so.
        """
        parser = FormParser()
        parser.feed(html_content)

        # Check for file inputs directly inside a form that do not reference
        # any form through the form attribute
        for form in parser.forms:
            if (
                form["file_form"]
                and form["form_attrs"].get("enctype") != "multipart/form-data"
                and not any(
                    elem.get("formenctype") == "multipart/form-data"
                    for elem in form["submit_element_attrs"]
                )
            ):
                if form_id := form["form_attrs"].get("id"):
                    alert = self.messages["form_id_missing_enctype"].format(
                        form_id=form_id
                    )
                else:
                    alert = self.messages["form_missing_enctype"]
                self.add_alert({"alert": alert})

        # Check for file inputs that reference a form
        form_attrs_by_id = {
            form["form_attrs"].get("id"): form["form_attrs"] for form in parser.forms
        }

        for attrs in parser.referenced_file_inputs:
            form_id = attrs.get("form")
            if form_id and attrs.get("type") == "file":
                form_attrs = form_attrs_by_id.get(form_id)
                if form_attrs and form_attrs.get("enctype") != "multipart/form-data":
                    alert = self.messages["input_refs_form_missing_enctype"].format(
                        form_id=form_id
                    )
                    self.add_alert({"alert": alert})

        return self.alerts

    def generate_stats(self, request, response):
        if is_processable_html_response(response):
            html_content = response.content.decode(response.charset)
            self.check_invalid_file_form_configuration(html_content)

        # Further alert checks can go here

        # Write all alerts to record_stats
        self.record_stats({"alerts": self.alerts})



import uuid

from django.test import TestCase

from debug_toolbar.models import HistoryEntry


class HistoryEntryTestCase(TestCase):
    def test_str_method(self):
        test_uuid = uuid.uuid4()
        entry = HistoryEntry(request_id=test_uuid)
        self.assertEqual(str(entry), str(test_uuid))

    def test_data_field_default(self):
        """Test that the data field defaults to an empty dict"""
        entry = HistoryEntry(request_id=uuid.uuid4())
        self.assertEqual(entry.data, {})

    def test_model_persistence(self):
        """Test saving and retrieving a model instance"""
        test_uuid = uuid.uuid4()
        entry = HistoryEntry(request_id=test_uuid, data={"test": True})
        entry.save()

        # Retrieve from database and verify
        saved_entry = HistoryEntry.objects.get(request_id=test_uuid)
        self.assertEqual(saved_entry.data, {"test": True})
        self.assertEqual(str(saved_entry), str(test_uuid))

    def test_default_ordering(self):
        """Test that the default ordering is by created_at in descending order"""
        self.assertEqual(HistoryEntry._meta.ordering, ["-created_at"])



from django.conf import settings
from django.db import models
from django.db.models import JSONField


class NonAsciiRepr:
    def __repr__(self):
        return "nôt åscíì"


class Binary(models.Model):
    field = models.BinaryField()

    def __str__(self):
        return ""


class PostgresJSON(models.Model):
    field = JSONField()

    def __str__(self):
        return ""


if settings.USE_GIS:
    from django.contrib.gis.db import models as gismodels

    class Location(gismodels.Model):
        point = gismodels.PointField()

        def __str__(self):
            return ""
